#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import octobot_trading.api as api
import octobot_trading.exchanges as exchanges

from tests.exchanges import exchange_manager
from tests import event_loop


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_build_trading_modes_if_required(exchange_manager):
    builder = exchanges.ExchangeBuilder({}, "binanceus")
    builder.exchange_manager = exchange_manager

    # no set trader: no trading mode creation attempt
    assert builder.exchange_manager.trader is None
    await builder._build_trading_modes_if_required(None)
    assert builder.exchange_manager.trader is None

    # with trader simulator: will attempt to create a trading mode and fail (because of the None arg)
    builder.is_simulated()
    with pytest.raises(AttributeError):
        await builder._build_trading_modes_if_required(None)


async def test_build_collector_exchange(exchange_manager):
    new_exchange_manager = await api.create_exchange_builder(exchange_manager.config, exchange_manager.exchange_name) \
        .is_simulated() \
        .is_rest_only() \
        .is_exchange_only() \
        .is_without_auth() \
        .is_ignoring_config() \
        .disable_trading_mode() \
        .use_tentacles_setup_config(exchange_manager.tentacles_setup_config) \
        .build()
    assert new_exchange_manager is not exchange_manager
    assert new_exchange_manager.config is exchange_manager.config
    assert new_exchange_manager.exchange_name == exchange_manager.exchange_name
    assert new_exchange_manager.tentacles_setup_config is exchange_manager.tentacles_setup_config
    await new_exchange_manager.stop()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders.groups import balanced_take_profit_and_stop_order_group
from octobot_trading.personal_data.orders.groups.balanced_take_profit_and_stop_order_group import (
    BalancedTakeProfitAndStopOrderGroup,
)

from octobot_trading.personal_data.orders.groups import one_cancels_the_other_order_group
from octobot_trading.personal_data.orders.groups.one_cancels_the_other_order_group import (
    OneCancelsTheOtherOrderGroup,
)


from octobot_trading.personal_data.orders.groups import group_util
from octobot_trading.personal_data.orders.groups.group_util import (
    get_group_class,
    get_or_create_order_group_from_storage_order_details,
)

__all__ = [
    "BalancedTakeProfitAndStopOrderGroup",
    "OneCancelsTheOtherOrderGroup",
    "get_group_class",
    "get_or_create_order_group_from_storage_order_details",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.exchanges.adapters as adapters


class ExchangeSimulatorAdapter(adapters.AbstractAdapter):

    def __init__(self, connector):
        super().__init__(connector)
        self._tentacle_adapter_proxy = None

    def _get_tentacle_adapter_proxy(self):
        if self._tentacle_adapter_proxy is None:
            return self
        return self._tentacle_adapter_proxy

    def set_tentacles_adapter_proxy(self, adapter_class):
        self._tentacle_adapter_proxy = adapter_class(self.connector)

    def adapt_market_status(self, raw, remove_price_limits=False, **kwargs):
        return self._get_tentacle_adapter_proxy().adapt_market_status(
            raw, remove_price_limits=remove_price_limits, **kwargs
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os

import pytest
import asyncio

import octobot_commons.constants as constants
from octobot_trading.enums import MarkPriceSources
import octobot_trading.constants
from octobot_trading.exchange_data.prices.prices_manager import calculate_mark_price_from_recent_trade_prices

from tests.exchanges import backtesting_exchange_manager, backtesting_config, fake_backtesting
from tests.exchange_data import price_events_manager, prices_manager
from tests import event_loop

# All test coroutines will be treated as marked.
from tests.test_utils.random_numbers import decimal_random_price

pytestmark = pytest.mark.asyncio


async def test_initialize(prices_manager):
    assert prices_manager.mark_price == octobot_trading.constants.ZERO
    assert prices_manager.mark_price_set_time == 0
    assert not prices_manager.valid_price_received_event.is_set()

    # should be reset in init
    prices_manager.mark_price = decimal.Decimal(10)
    prices_manager.mark_price_set_time = 10
    prices_manager.valid_price_received_event.set()

    await prices_manager.initialize()
    assert prices_manager.mark_price == octobot_trading.constants.ZERO
    assert prices_manager.mark_price_set_time == 0
    assert not prices_manager.valid_price_received_event.is_set()


async def test_set_mark_price(prices_manager):
    prices_manager.set_mark_price(decimal.Decimal(10), MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    assert prices_manager.mark_price == decimal.Decimal(10)
    check_event_is_set(prices_manager)


async def test_set_mark_price_for_exchange_source(prices_manager):
    prices_manager.set_mark_price(decimal.Decimal(10), MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    assert prices_manager.mark_price == decimal.Decimal(10)
    check_event_is_set(prices_manager)
    prices_manager.set_mark_price(decimal.Decimal(25), MarkPriceSources.RECENT_TRADE_AVERAGE.value)
    assert prices_manager.mark_price == decimal.Decimal(10)  # Drop first RT update
    prices_manager.set_mark_price(decimal.Decimal(30), MarkPriceSources.RECENT_TRADE_AVERAGE.value)
    assert prices_manager.mark_price == decimal.Decimal(30)
    prices_manager.set_mark_price(decimal.Decimal(20), MarkPriceSources.TICKER_CLOSE_PRICE.value)
    assert prices_manager.mark_price == decimal.Decimal(30)
    prices_manager.set_mark_price(decimal.Decimal(15), MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    assert prices_manager.mark_price == decimal.Decimal(15)


async def test_set_mark_price_for_ticker_source_only(prices_manager):
    assert prices_manager.set_mark_price(decimal.Decimal(10), MarkPriceSources.TICKER_CLOSE_PRICE.value) is True
    assert prices_manager.mark_price == decimal.Decimal(10)
    check_event_is_set(prices_manager)


async def test_set_mark_price_for_rt_source_only(prices_manager):
    assert prices_manager.set_mark_price(decimal.Decimal(10), MarkPriceSources.RECENT_TRADE_AVERAGE.value) is False
    assert prices_manager.mark_price == decimal.Decimal(0)  # drop first RT mark price
    assert prices_manager.set_mark_price(decimal.Decimal(25), MarkPriceSources.RECENT_TRADE_AVERAGE.value) is True
    assert prices_manager.mark_price == decimal.Decimal(25)
    check_event_is_set(prices_manager)


async def test_set_mark_price_for_ticker_with_rt_source_outdated(prices_manager):
    prices_manager.set_mark_price(decimal.Decimal(5), MarkPriceSources.RECENT_TRADE_AVERAGE.value)
    assert prices_manager.mark_price == decimal.Decimal(0)  # Drop first RT update
    assert prices_manager.set_mark_price(decimal.Decimal(10), MarkPriceSources.RECENT_TRADE_AVERAGE.value) is True
    assert prices_manager.mark_price == decimal.Decimal(10)
    check_event_is_set(prices_manager)
    assert prices_manager.set_mark_price(decimal.Decimal(40), MarkPriceSources.TICKER_CLOSE_PRICE.value) is False
    assert prices_manager.mark_price == decimal.Decimal(10)  # should not be updated because RECENT_TRADE_AVERAGE has not expired
    check_event_is_set(prices_manager)
    assert prices_manager.set_mark_price(decimal.Decimal(20), MarkPriceSources.TICKER_CLOSE_PRICE.value) is False
    assert prices_manager.mark_price == decimal.Decimal(10)  # should not be updated because RECENT_TRADE_AVERAGE has not expired
    check_event_is_set(prices_manager)
    # force rt source expiration
    if not os.getenv('CYTHON_IGNORE'):
        prices_manager.mark_price_from_sources[MarkPriceSources.RECENT_TRADE_AVERAGE.value] = None
        assert prices_manager.set_mark_price(decimal.Decimal(40), MarkPriceSources.TICKER_CLOSE_PRICE.value) is True
        assert prices_manager.mark_price == decimal.Decimal(40)  # should be updated because RECENT_TRADE_AVERAGE has expired
        check_event_is_set(prices_manager)
        assert prices_manager.set_mark_price(decimal.Decimal(8), MarkPriceSources.RECENT_TRADE_AVERAGE.value) is False
        assert prices_manager.mark_price == decimal.Decimal(40)  # Drop first RT update after reset
        assert prices_manager.mark_price_from_sources[MarkPriceSources.RECENT_TRADE_AVERAGE.value][1] == 0
        assert prices_manager.set_mark_price(decimal.Decimal(15), MarkPriceSources.RECENT_TRADE_AVERAGE.value) is True
        assert prices_manager.mark_price == decimal.Decimal(15)
        assert prices_manager.mark_price_from_sources[MarkPriceSources.RECENT_TRADE_AVERAGE.value][1] != 0
        check_event_is_set(prices_manager)
        assert prices_manager.set_mark_price(decimal.Decimal(20), MarkPriceSources.TICKER_CLOSE_PRICE.value) is False
        assert prices_manager.mark_price == decimal.Decimal(15)  # should not be updated because RECENT_TRADE_AVERAGE has not expired
        check_event_is_set(prices_manager)

        # force rt source expiration
        prices_manager.mark_price_from_sources[MarkPriceSources.RECENT_TRADE_AVERAGE.value] = (
            prices_manager.mark_price_from_sources[MarkPriceSources.RECENT_TRADE_AVERAGE.value][0],
            prices_manager.mark_price_from_sources[MarkPriceSources.RECENT_TRADE_AVERAGE.value][1] -
            5 * constants.MINUTE_TO_SECONDS
        )
        prices_manager.set_mark_price(decimal.Decimal(40), MarkPriceSources.TICKER_CLOSE_PRICE.value)
        assert prices_manager.mark_price == decimal.Decimal(40)  # should be updated because RECENT_TRADE_AVERAGE has expired
        check_event_is_set(prices_manager)


async def test_get_mark_price(prices_manager):
    # without a set price
    with pytest.raises(asyncio.TimeoutError):
        await prices_manager.get_mark_price(0.01)
    assert not prices_manager.valid_price_received_event.is_set()

    # set price
    prices_manager.set_mark_price(decimal.Decimal(10), MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    assert await prices_manager.get_mark_price(0.01) == decimal.Decimal(10)
    assert prices_manager.valid_price_received_event.is_set()

    # expired price
    if not os.getenv('CYTHON_IGNORE'):
        prices_manager.exchange_manager.backtesting.time_manager.current_timestamp = 66666666
        with pytest.raises(asyncio.TimeoutError):
            await prices_manager.get_mark_price(0.01)
        assert not prices_manager.valid_price_received_event.is_set()

    # reset price with this time
    prices_manager.set_mark_price(decimal.Decimal(10), MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    assert await prices_manager.get_mark_price(0.01) == decimal.Decimal(10)
    assert prices_manager.valid_price_received_event.is_set()

    # current time move within allowed range
    if not os.getenv('CYTHON_IGNORE'):
        prices_manager.exchange_manager.backtesting.time_manager.current_timestamp = 1
        assert await prices_manager.get_mark_price(0.01) == decimal.Decimal(10)
        assert prices_manager.valid_price_received_event.is_set()

    # new value
    prices_manager.set_mark_price(decimal.Decimal(42.0000172), MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    assert await prices_manager.get_mark_price(0.01) == decimal.Decimal(42.0000172)
    assert prices_manager.valid_price_received_event.is_set()

    # random value
    random_mark_price = decimal_random_price()
    prices_manager.set_mark_price(random_mark_price, MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    assert await prices_manager.get_mark_price(0.01) == random_mark_price
    assert prices_manager.valid_price_received_event.is_set()


async def test_calculate_mark_price_from_recent_trade_prices():
    assert calculate_mark_price_from_recent_trade_prices([10, 5, 7]) == decimal.Decimal('7.333333333333333333333333333')
    assert calculate_mark_price_from_recent_trade_prices([10, 20]) == decimal.Decimal(15)
    assert calculate_mark_price_from_recent_trade_prices([]) == decimal.Decimal(0)


def check_event_is_set(prices_manager):
    if not os.getenv('CYTHON_IGNORE'):
        assert prices_manager.mark_price_set_time == prices_manager.exchange_manager.exchange.get_exchange_current_time()
        assert prices_manager.valid_price_received_event.is_set()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import pytest_asyncio

import octobot_commons.constants as commons_constants
import octobot_commons.symbols.symbol_util as symbol_util
from octobot_commons.tests.test_config import load_test_config
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.exchange_data.contracts as contracts
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator
from octobot_trading.exchanges.traders.trader import Trader


async def create_trader_from_exchange_manager(exchange_manager, simulated=False, contract=None):
    config = load_test_config()
    if simulated:
        config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_ENABLED_OPTION] = True
        trader_inst = TraderSimulator(load_test_config(), exchange_manager)
    else:
        config[commons_constants.CONFIG_TRADER][commons_constants.CONFIG_ENABLED_OPTION] = True
        trader_inst = Trader(load_test_config(), exchange_manager)
    await trader_inst.initialize()
    if contract is not None:
        return config, exchange_manager, trader_inst, contract
    return config, exchange_manager, trader_inst


@pytest_asyncio.fixture
async def trader(exchange_manager):
    return await create_trader_from_exchange_manager(exchange_manager)


@pytest_asyncio.fixture
async def margin_trader(margin_exchange_manager):
    return await create_trader_from_exchange_manager(margin_exchange_manager)


@pytest_asyncio.fixture
async def future_trader(future_exchange_manager):
    return await create_trader_from_exchange_manager(future_exchange_manager)


@pytest_asyncio.fixture
async def trader_simulator(simulated_exchange_manager):
    return await create_trader_from_exchange_manager(simulated_exchange_manager, simulated=True)


@pytest_asyncio.fixture
async def margin_trader_simulator(margin_simulated_exchange_manager):
    return await create_trader_from_exchange_manager(margin_simulated_exchange_manager, simulated=True)


DEFAULT_FUTURE_SYMBOL = str(symbol_util.parse_symbol("BTC/USDT:USDT"))
DEFAULT_FUTURE_FUNDING_RATE = decimal.Decimal('0.01')
DEFAULT_FUTURE_SYMBOL_LEVERAGE = constants.ONE
DEFAULT_FUTURE_SYMBOL_MAX_LEVERAGE = constants.ONE_HUNDRED
DEFAULT_FUTURE_SYMBOL_MARGIN_TYPE = enums.MarginType.ISOLATED


def get_default_future_inverse_contract():
    return contracts.FutureContract(
        pair=DEFAULT_FUTURE_SYMBOL,
        margin_type=DEFAULT_FUTURE_SYMBOL_MARGIN_TYPE,
        contract_type=enums.FutureContractType.INVERSE_PERPETUAL,
        current_leverage=DEFAULT_FUTURE_SYMBOL_LEVERAGE,
        maximum_leverage=DEFAULT_FUTURE_SYMBOL_MAX_LEVERAGE)


def get_default_future_linear_contract():
    return contracts.FutureContract(
        pair=DEFAULT_FUTURE_SYMBOL,
        margin_type=DEFAULT_FUTURE_SYMBOL_MARGIN_TYPE,
        contract_type=enums.FutureContractType.LINEAR_PERPETUAL,
        current_leverage=DEFAULT_FUTURE_SYMBOL_LEVERAGE,
        maximum_leverage=DEFAULT_FUTURE_SYMBOL_MAX_LEVERAGE)


@pytest_asyncio.fixture
async def future_trader_simulator_with_default_inverse(future_simulated_exchange_manager):
    contract = get_default_future_inverse_contract()
    future_simulated_exchange_manager.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, contract)
    return await create_trader_from_exchange_manager(future_simulated_exchange_manager,
                                                     simulated=True,
                                                     contract=contract)


@pytest_asyncio.fixture
async def future_trader_simulator_with_default_linear(future_simulated_exchange_manager):
    contract = get_default_future_linear_contract()
    future_simulated_exchange_manager.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, contract)
    return await create_trader_from_exchange_manager(future_simulated_exchange_manager,
                                                     simulated=True,
                                                     contract=contract)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import enum


class QuantityType(enum.Enum):
    DELTA = ""
    DELTA_EXPLICIT = "d"
    DELTA_QUOTE = "q"
    DELTA_BASE = "b"
    PERCENT = "%"
    AVAILABLE = "a"
    POSITION = "p"
    ENTRY = "e"
    AVAILABLE_PERCENT = "a%"
    POSITION_PERCENT = "p%"
    ENTRY_PERCENT = "e%"
    CURRENT_SYMBOL_ASSETS_PERCENT = "s%"
    TRADED_SYMBOLS_ASSETS_PERCENT = "t%"
    FLAT = "@"
    UNKNOWN = "?"

    @staticmethod
    def parse(value):
        try:
            # try reading directly as enum
            return QuantityType(value), value
        except ValueError:
            # try with letters in reverse order
            reversed_value = value[::-1]
            return QuantityType(reversed_value), value



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_commons.symbols as symbols


class TradePnl:
    def __init__(self, entries, closes):
        self.entries = entries
        self.closes = closes

    def get_entry_time(self) -> float:
        try:
            return min(
                entry.executed_time
                for entry in self.entries
            )
        except ValueError as err:
            raise errors.IncompletePNLError from err

    def get_close_time(self) -> float:
        try:
            return max(
                close.executed_time
                for close in self.closes
            )
        except ValueError as err:
            raise errors.IncompletePNLError from err

    def get_total_entry_quantity(self) -> decimal.Decimal:
        return sum(
            entry.executed_quantity
            for entry in self.entries
        ) or constants.ZERO

    def get_total_close_quantity(self) -> decimal.Decimal:
        return sum(
            close.executed_quantity
            for close in self.closes
        ) or constants.ZERO

    def get_entry_price(self) -> decimal.Decimal:
        try:
            return sum(
                entry.executed_price
                for entry in self.entries
            ) / len(self.entries)
        except ZeroDivisionError as err:
            raise errors.IncompletePNLError from err

    def get_close_price(self) -> decimal.Decimal:
        try:
            return sum(
                close.executed_price
                for close in self.closes
            ) / len(self.closes)
        except ZeroDivisionError as err:
            raise errors.IncompletePNLError from err

    def get_entry_total_cost(self):
        return self.get_entry_price() * self.get_total_entry_quantity()

    def get_close_total_cost(self):
        return self.get_close_price() * self.get_total_close_quantity()

    def get_close_ratio(self):
        try:
            if self.entries[0].side is enums.TradeOrderSide.BUY:
                return min(self.get_total_close_quantity() / self.get_total_entry_quantity(), constants.ONE)
            return min(self.get_close_total_cost() / self.get_entry_total_cost(), constants.ONE)
        except (IndexError, decimal.DivisionByZero, decimal.InvalidOperation) as err:
            raise errors.IncompletePNLError from err

    def get_closed_entry_value(self) -> decimal.Decimal:
        return self.get_entry_price() * self.get_total_entry_quantity()

    def get_closed_close_value(self) -> decimal.Decimal:
        return self.get_close_price() * self.get_closed_pnl_quantity()

    def get_closed_pnl_quantity(self):
        try:
            if self.entries[0].side is enums.TradeOrderSide.BUY:
                # entry is a buy, exit is a sell: take exit size capped at the entry size
                # (can't account in pnl for more than what has been bought)
                return min(self.get_total_close_quantity(), self.get_total_entry_quantity())
            # entry is a sell, exit is a buy: take closing size capped at the equivalent entry cost
            # (can't account in pnl for more than what has been sold for entry)
            entry_cost = self.get_entry_price() * self.get_total_entry_quantity()
            max_close_quantity = entry_cost / self.get_close_price()
            return min(self.get_total_close_quantity(), max_close_quantity)
        except IndexError as err:
            raise errors.IncompletePNLError from err

    def _get_fees(self, trade) -> decimal.Decimal:
        if not trade.fee:
            return constants.ZERO
        symbol = symbols.parse_symbol(trade.symbol)
        # return fees denominated in quote
        fees = order_util.get_fees_for_currency(trade.fee, symbol.quote)
        return fees \
            + order_util.get_fees_for_currency(trade.fee, symbol.base) * trade.executed_price

    def get_paid_special_fees_by_currency(self) -> dict:
        """
        :return: a dict containing fees paid in currencies different from the base or quote of the trades pair
        values are not converted into base currency of the trading pair
        """
        if not self.entries:
            return {}
        try:
            fees = {}
            base_and_quote = symbols.parse_symbol(self.entries[0].symbol).base_and_quote()
            for trade in (*self.entries, *self.closes):
                if trade.fee:
                    currency = trade.fee[enums.FeePropertyColumns.CURRENCY.value]
                    if currency in base_and_quote:
                        # not a special fee
                        continue
                    if currency in fees:
                        fees[currency] += order_util.get_fees_for_currency(trade.fee, currency)
                    else:
                        fees[currency] = order_util.get_fees_for_currency(trade.fee, currency)
            return fees
        except IndexError as err:
            raise errors.IncompletePNLError from err

    def get_paid_regular_fees_in_quote(self) -> decimal.Decimal:
        """
        :return: the total value (in quote) of paid fees when paid in base or quote of the trades pair
        """
        return sum(
            self._get_fees(trade)
            for trade in (*self.entries, *self.closes)
        ) or constants.ZERO

    def get_profits(self) -> (decimal.Decimal, decimal.Decimal):
        """
        :return: the pnl profits as flat value and percent
        """
        close_holdings = self.get_closed_close_value() - self.get_paid_regular_fees_in_quote()
        entry_holdings = self.get_closed_entry_value() * self.get_close_ratio()
        try:
            percent_profit = (close_holdings * constants.ONE_HUNDRED / entry_holdings) - constants.ONE_HUNDRED
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            percent_profit = constants.ZERO
        return (
            close_holdings - entry_holdings,
            percent_profit
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging

import octobot_trading.personal_data as personal_data
import octobot_trading.enums
import octobot_trading.constants as constants

LOGGER = logging.get_logger(constants.API_LOGGER_TAG)


async def create_order(exchange_manager,
                       order_type: octobot_trading.enums.TraderOrderType,
                       symbol: str,
                       current_price: float,
                       quantity: float,
                       price: float,
                       wait_for_creation=True,
                       creation_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT) -> personal_data.Order:
    return await exchange_manager.trader.create_order(
        exchange_manager.trader.create_order_instance(order_type=order_type,
                                                      symbol=symbol,
                                                      current_price=current_price,
                                                      quantity=quantity,
                                                      price=price),
        wait_for_creation=wait_for_creation,
        creation_timeout=creation_timeout
    )


def get_open_orders(exchange_manager, symbol=None) -> list:
    return exchange_manager.exchange_personal_data.orders_manager.get_open_orders(symbol=symbol)


async def cancel_all_open_orders(exchange_manager, emit_trading_signals=True) -> bool:
    return await exchange_manager.trader.cancel_all_open_orders(emit_trading_signals=emit_trading_signals)


async def cancel_all_open_orders_with_currency(exchange_manager, currency, emit_trading_signals=True) -> bool:
    return await exchange_manager.trader.cancel_all_open_orders_with_currency(currency, emit_trading_signals=emit_trading_signals)


async def cancel_order_with_id(exchange_manager, order_id, emit_trading_signals=True, wait_for_cancelling=True) -> bool:
    return await exchange_manager.trader.cancel_order_with_id(order_id, emit_trading_signals=emit_trading_signals,
                                                              wait_for_cancelling=wait_for_cancelling)


def get_order_exchange_name(order) -> str:
    return order.exchange_manager.get_exchange_name()


def order_to_dict(order) -> dict:
    return order.to_dict()


def parse_order_type(dict_order) -> octobot_trading.enums.TraderOrderType:
    return personal_data.parse_order_type(dict_order)[1]


def parse_order_status(dict_order) -> octobot_trading.enums.OrderStatus:
    return personal_data.parse_order_status(dict_order)


def get_order_profitability(exchange_manager, order_id) -> float:
    try:
        return exchange_manager.exchange_personal_data.orders_manager.get_order(order_id).get_profitability()
    except KeyError:
        # try in trades (order might be filled and stored in trades)
        return exchange_manager.exchange_personal_data.trades_manager.get_trade(order_id).trade_profitability


def get_minimal_order_cost(exchange_manager, symbol) -> float:
    return personal_data.get_minimal_order_cost(
        exchange_manager.exchange.get_market_status(symbol, with_fixer=False)
    )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.enums as enums


def get_positions(exchange_manager) -> list:
    return exchange_manager.exchange_personal_data.positions_manager.get_symbol_positions()


async def close_position(exchange_manager, symbol: str, side: enums.PositionSide,
                         limit_price: decimal.Decimal = None, emit_trading_signals=True) -> int:
    for position in exchange_manager.exchange_personal_data.positions_manager.get_symbol_positions(symbol):
        if position.side is side:
            if position.is_idle():
                return 0
            return 1 if await exchange_manager.trader.close_position(
                position,
                limit_price=limit_price,
                emit_trading_signals=emit_trading_signals
            ) else 0
    return 0



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock

import trading_backend.exchanges
import octobot_commons.constants as commons_constants
import octobot_commons.configuration as commons_configuration
import octobot_trading.constants as constants
import octobot_trading.exchanges as exchanges
import octobot_trading.exchanges.util.exchange_util as exchange_util

from tests import event_loop
from tests.exchanges import MockedRestExchange, MockedAutoFillRestExchange
import octobot_tentacles_manager.api as api


@pytest.fixture
def tentacles_setup_config():
    setup_config = mock.Mock()
    setup_config.is_tentacle_activated = mock.Mock(return_value=True)
    return setup_config


@pytest.fixture()
def exchange_config():
    return {
        commons_constants.CONFIG_EXCHANGE_KEY: commons_configuration.encrypt("01234").decode(),
        commons_constants.CONFIG_EXCHANGE_SECRET: commons_configuration.encrypt("012345").decode()
    }


@pytest.fixture()
def supported_exchanges():
    return {
        "plop.exchange": exchanges.ExchangeDetails("id_plop", "name_plop", "url_plop", "api_plop", "logo_plop", True),
        "blip": exchanges.ExchangeDetails("id_blip", "name_blip", "url_blip", "api_blip", "logo_blip", True)
    }


@pytest.mark.asyncio
async def test_is_compatible_account_with_checked_exchange(exchange_config, tentacles_setup_config):
    with mock.patch.object(trading_backend.exchanges.Huobi, "is_valid_account",
                           mock.AsyncMock(return_value=(True, None))) as is_valid_account_mock:
        compatible, auth, error = await exchanges.is_compatible_account("huobi", exchange_config,
                                                                        tentacles_setup_config, False)
        assert compatible is True
        assert auth is True
        assert error is None
        is_valid_account_mock.assert_called_once()
    with mock.patch.object(trading_backend.exchanges.Huobi, "is_valid_account",
                           mock.AsyncMock(return_value=(False, "plop"))) as is_valid_account_mock:
        compatible, auth, error = await exchanges.is_compatible_account("huobi", exchange_config,
                                                                        tentacles_setup_config, False)
        # still True as on spot trading
        assert compatible is True
        assert auth is True
        assert error is None
        is_valid_account_mock.assert_called_once()
    exchange_config[commons_constants.CONFIG_EXCHANGE_TYPE] = commons_constants.CONFIG_EXCHANGE_FUTURE
    with mock.patch.object(trading_backend.exchanges.Huobi, "is_valid_account",
                           mock.AsyncMock(return_value=(False, "plop"))) as is_valid_account_mock:
        compatible, auth, error = await exchanges.is_compatible_account("huobi", exchange_config,
                                                                        tentacles_setup_config, False)
        assert compatible is False
        assert auth is True
        assert "plop" in error and len(error) > len("plop")
        is_valid_account_mock.assert_called_once()


def test_get_partners_explanation_message():
    assert ".cloud/en/guides" in exchanges.get_partners_explanation_message()


def test_log_time_sync_error():
    logger = mock.Mock()
    exchanges.log_time_sync_error(logger, "exchange_name", "error", "hello call")
    args = logger.error.call_args[0][0]
    assert "exchange_name".capitalize() in args
    assert "error" in args
    assert "hello call" in args
    assert ".cloud/en/guides" in args


@pytest.mark.asyncio
async def test_is_compatible_account_with_unchecked_exchange(exchange_config, tentacles_setup_config):
    compatible, auth, error = await exchanges.is_compatible_account("hitbtc", exchange_config, tentacles_setup_config,
                                                                    False)
    assert compatible is False
    assert auth is False
    assert isinstance(error, str)
    exchange_config[commons_constants.CONFIG_EXCHANGE_TYPE] = commons_constants.CONFIG_EXCHANGE_FUTURE
    with mock.patch.object(trading_backend.exchanges.Exchange, "is_valid_account",
                           mock.AsyncMock(return_value=(True, "plop"))) as is_valid_account_mock:
        compatible, auth, error = await exchanges.is_compatible_account("hitbtc", exchange_config,
                                                                        tentacles_setup_config, False)
        assert compatible is True
        assert auth is True
        assert "plop" in error and len(error) > len("plop")
        is_valid_account_mock.assert_called_once()
    with mock.patch.object(trading_backend.exchanges.Exchange, "is_valid_account",
                           mock.AsyncMock(side_effect=trading_backend.ExchangeAuthError)) as is_valid_account_mock:
        compatible, auth, error = await exchanges.is_compatible_account("hitbtc", exchange_config,
                                                                        tentacles_setup_config, False)
        assert compatible is False
        assert auth is False
        assert "authentication" in error and len(error) > len("authentication")
        is_valid_account_mock.assert_called_once()


def test_get_auto_filled_exchange_names(tentacles_setup_config, supported_exchanges):
    with mock.patch.object(api, "get_tentacle_config", mock.Mock()) as get_tentacle_config_mock:
        # no auto filled exchanges
        assert exchanges.get_auto_filled_exchange_names(tentacles_setup_config) == []
        get_tentacle_config_mock.assert_called_once_with(tentacles_setup_config, MockedAutoFillRestExchange)
        get_tentacle_config_mock.reset_mock()

        with MockedAutoFillRestExchange.patched_supported_exchanges(supported_exchanges):
            auto_filled_exchanges = exchanges.get_auto_filled_exchange_names(tentacles_setup_config)
            assert auto_filled_exchanges == list(supported_exchanges)
            assert "blip" in auto_filled_exchanges
            get_tentacle_config_mock.assert_called_once_with(tentacles_setup_config, MockedAutoFillRestExchange)


def test_get_exchange_class_from_name(tentacles_setup_config, supported_exchanges):
    # not found exchange
    assert exchanges.get_exchange_class_from_name(
        exchanges.RestExchange, "plop", tentacles_setup_config, True,
        strict_name_matching=False
    ) == exchanges.DefaultRestExchange
    with mock.patch.object(api, "get_tentacle_config", mock.Mock()) as get_tentacle_config_mock:
        assert exchanges.get_exchange_class_from_name(
            exchanges.RestExchange, "plop", tentacles_setup_config, False,
            strict_name_matching=False
        ) == MockedRestExchange
        get_tentacle_config_mock.assert_not_called()
        assert exchanges.get_exchange_class_from_name(
            exchanges.RestExchange, "plop", tentacles_setup_config, True,
            strict_name_matching=True
        ) is None
        get_tentacle_config_mock.assert_called_once()
        get_tentacle_config_mock.reset_mock()

    # regular exchange
    assert exchanges.get_exchange_class_from_name(
        exchanges.RestExchange, MockedRestExchange.get_name(), tentacles_setup_config, True,
        strict_name_matching=False
    ) == exchanges.DefaultRestExchange
    assert exchanges.get_exchange_class_from_name(
        exchanges.RestExchange, MockedRestExchange.get_name(), tentacles_setup_config, False,
        strict_name_matching=False
    ) == MockedRestExchange
    assert exchanges.get_exchange_class_from_name(
        exchanges.RestExchange, MockedRestExchange.get_name(), tentacles_setup_config, True,
        strict_name_matching=True
    ) == MockedRestExchange

    with mock.patch.object(api, "get_tentacle_config", mock.Mock()) as get_tentacle_config_mock:
        # auto-filled exchange
        with MockedAutoFillRestExchange.patched_supported_exchanges(supported_exchanges):
            assert exchanges.get_exchange_class_from_name(
                exchanges.RestExchange, MockedRestExchange.get_name(), tentacles_setup_config, True,
                strict_name_matching=False
            ) == exchanges.DefaultRestExchange
            assert exchanges.get_exchange_class_from_name(
                exchanges.RestExchange, MockedRestExchange.get_name(), tentacles_setup_config, False,
                strict_name_matching=False
            ) == MockedRestExchange
            assert exchanges.get_exchange_class_from_name(
                exchanges.RestExchange, MockedRestExchange.get_name(), tentacles_setup_config, True,
                strict_name_matching=True
            ) == MockedRestExchange

            get_tentacle_config_mock.assert_not_called()

            assert exchanges.get_exchange_class_from_name(
                exchanges.RestExchange, "blip", tentacles_setup_config, True,
                strict_name_matching=False
            ) == exchanges.DefaultRestExchange
            assert exchanges.get_exchange_class_from_name(
                exchanges.RestExchange, "blip", tentacles_setup_config, False,
                strict_name_matching=False
            ) == MockedRestExchange
            get_tentacle_config_mock.assert_not_called()
            assert exchanges.get_exchange_class_from_name(
                exchanges.RestExchange, "blip", tentacles_setup_config, True,
                strict_name_matching=True
            ) == MockedAutoFillRestExchange
            get_tentacle_config_mock.assert_called_once()


@pytest.mark.asyncio
async def test_get_exchange_details(tentacles_setup_config, supported_exchanges):
    with mock.patch.object(api, "get_tentacle_config", mock.Mock()) as get_tentacle_config_mock:
        # not found exchange
        with pytest.raises(KeyError):
            await exchanges.get_exchange_details(
                "blip", False, tentacles_setup_config, None
            )
        get_tentacle_config_mock.assert_not_called()
        with pytest.raises(KeyError):
            await exchanges.get_exchange_details(
                "blip", True, tentacles_setup_config, None
            )
        get_tentacle_config_mock.assert_called_once()
        get_tentacle_config_mock.reset_mock()

        # regular exchange
        details = await exchanges.get_exchange_details(
            "binance", False, tentacles_setup_config, None
        )
        assert details.id == "binance"
        assert details.name == "Binance"
        assert details.url == "https://www.binance.com"
        assert len(details.api) > 1
        assert "https://user-images.githubusercontent.com/" in details.logo_url
        assert details.has_websocket is False   # default value
        get_tentacle_config_mock.assert_not_called()

        # auto-filled exchange
        with MockedAutoFillRestExchange.patched_supported_exchanges(supported_exchanges):
            with pytest.raises(KeyError):
                await exchanges.get_exchange_details(
                    "blip", False, tentacles_setup_config, None
                )
            get_tentacle_config_mock.assert_not_called()
            details = await exchanges.get_exchange_details(
                "blip", True, tentacles_setup_config, None
            )
            assert details == supported_exchanges["blip"]
            get_tentacle_config_mock.assert_called_once()


def test_is_api_permission_error():
    with mock.patch.object(exchange_util, "is_error_on_this_type") as is_error_on_this_type_mock:
        err = Exception("plop")
        exchanges.is_api_permission_error(err)
        is_error_on_this_type_mock.assert_called_once_with(
            err, constants.EXCHANGE_PERMISSION_ERRORS
        )


def test_is_exchange_rules_compliancy_error():
    with mock.patch.object(exchange_util, "is_error_on_this_type") as is_error_on_this_type_mock:
        err = Exception("plop")
        exchanges.is_exchange_rules_compliancy_error(err)
        is_error_on_this_type_mock.assert_called_once_with(
            err, constants.EXCHANGE_COMPLIANCY_ERRORS
        )


def test_is_error_on_this_type():
    assert exchange_util.is_error_on_this_type(Exception("plop"), constants.EXCHANGE_PERMISSION_ERRORS) is False
    assert exchange_util.is_error_on_this_type(Exception("api key"), constants.EXCHANGE_PERMISSION_ERRORS) is True
    assert exchange_util.is_error_on_this_type(Exception("api"), constants.EXCHANGE_PERMISSION_ERRORS) is False
    assert exchange_util.is_error_on_this_type(Exception("api"), constants.EXCHANGE_COMPLIANCY_ERRORS) is False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import pytest
from octobot_commons.tests.test_config import load_test_config

from tests import event_loop
import octobot_trading.personal_data as personal_data
import octobot_trading.storage.orders_storage as orders_storage
from octobot_trading.enums import TradeOrderSide, TradeOrderType, TraderOrderType, StoredOrdersAttr
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator
from octobot_trading.api.exchange import cancel_ccxt_throttle_task

pytestmark = pytest.mark.asyncio


class TestOrderFactory:
    DEFAULT_SYMBOL = "BTC/USDT"
    EXCHANGE_MANAGER_CLASS_STRING = "binanceus"

    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, TestOrderFactory.EXCHANGE_MANAGER_CLASS_STRING)
        await exchange_manager.initialize()

        trader = TraderSimulator(config, exchange_manager)
        await trader.initialize()

        return config, exchange_manager, trader

    @staticmethod
    async def stop(exchange_manager):
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_parse_order_type(self):
        _, exchange_manager, trader_inst = await self.init_default()

        order_to_test = personal_data.Order(trader_inst)
        assert order_to_test.simulated is True

        ccxt_order_buy_market = {
            "side": TradeOrderSide.BUY,
            "type": TradeOrderType.MARKET
        }

        order_to_test.update_from_raw(ccxt_order_buy_market)
        assert order_to_test.order_type == TraderOrderType.BUY_MARKET

        ccxt_order_buy_limit = {
            "side": TradeOrderSide.BUY,
            "type": TradeOrderType.LIMIT
        }
        assert personal_data.parse_order_type(ccxt_order_buy_limit) == (TradeOrderSide.BUY, TraderOrderType.BUY_LIMIT)

        ccxt_order_sell_market = {
            "side": TradeOrderSide.SELL,
            "type": TradeOrderType.MARKET
        }
        assert personal_data.parse_order_type(ccxt_order_sell_market) == (TradeOrderSide.SELL, TraderOrderType.SELL_MARKET)

        ccxt_order_sell_limit = {
            "side": TradeOrderSide.SELL,
            "type": TradeOrderType.LIMIT
        }
        assert personal_data.parse_order_type(ccxt_order_sell_limit) == (TradeOrderSide.SELL, TraderOrderType.SELL_LIMIT)

        await self.stop(exchange_manager)

    async def test_create_order_from_dict(self):
        price = decimal.Decimal("100")
        quantity = decimal.Decimal("2")
        _, exchange_manager, trader_inst = await self.init_default()
        limit_order = personal_data.create_order_instance(
            trader_inst,
            TraderOrderType.SELL_LIMIT,
            self.DEFAULT_SYMBOL,
            price,
            quantity,
            price=price,
            order_id="123",
            tag="tag",
            reduce_only=True,
            exchange_creation_params={"plop": 1, "fake_param": True},
            associated_entry_id="1",
        )
        order_dict = limit_order.to_dict()
        created_from_dict = personal_data.create_order_from_dict(trader_inst, order_dict)
        assert created_from_dict.origin_price == limit_order.origin_price == price
        assert created_from_dict.origin_quantity == limit_order.origin_quantity == quantity
        assert created_from_dict.__class__ is limit_order.__class__ == personal_data.SellLimitOrder
        assert created_from_dict.symbol == limit_order.symbol == self.DEFAULT_SYMBOL
        assert created_from_dict.order_id == limit_order.order_id == "123"
        assert created_from_dict.tag == limit_order.tag == "tag"
        assert created_from_dict.reduce_only is limit_order.reduce_only is True
        # exchange_creation_params are not copied
        assert created_from_dict.exchange_creation_params == {}
        assert limit_order.exchange_creation_params == {"plop": 1, "fake_param": True}
        # associated_entry_ids are not copied
        assert created_from_dict.associated_entry_ids is None
        assert limit_order.associated_entry_ids == ["1"]
        await self.stop(exchange_manager)

    async def test_create_order_from_order_storage_details_with_simple_order(self):
        _, exchange_manager, trader_inst = await self.init_default()
        try:
        
            order = personal_data.BuyLimitOrder(trader_inst)
            order.update(order_type=TraderOrderType.BUY_LIMIT,
                         symbol="BTC/USDT",
                         current_price=decimal.Decimal("70"),
                         quantity=decimal.Decimal("10"),
                         price=decimal.Decimal("70"))
            order_storage_details = orders_storage._format_order(order, exchange_manager)
            order_storage_details[StoredOrdersAttr.ENTRIES.value] = ["11111"]

            pending_groups = {}
            created_order = await personal_data.create_order_from_order_storage_details(
                order_storage_details, exchange_manager, pending_groups
            )
            assert pending_groups == {}

            assert created_order.exchange_manager is exchange_manager
            assert created_order.origin_quantity == order.origin_quantity
            assert created_order.timestamp == order.timestamp
            assert created_order.creation_time == order.creation_time
            assert created_order.origin_price == order.origin_price
            assert created_order.__class__ is order.__class__
            # associated_entry_ids are added from order_storage_details but not in original order
            assert created_order.associated_entry_ids == ["11111"]
            assert order.associated_entry_ids is None

            # updated creation_time (as with chained orders): creation_time is used to restore order
            assert created_order.creation_time != 123
            order.creation_time = 123

            order_storage_details = orders_storage._format_order(order, exchange_manager)
            created_order = await personal_data.create_order_from_order_storage_details(
                order_storage_details, exchange_manager, pending_groups
            )
            assert pending_groups == {}
            assert created_order.exchange_manager is exchange_manager
            assert created_order.origin_quantity == order.origin_quantity
            assert created_order.timestamp == 123   # aligned with creation time
            assert created_order.creation_time == 123   # aligned with creation time
            assert created_order.origin_price == order.origin_price
            assert created_order.__class__ is order.__class__
            # associated_entry_ids are added from order_storage_details but not in original order
            assert created_order.associated_entry_ids is None
            assert order.associated_entry_ids is None

        finally:
            await self.stop(exchange_manager)
    
    async def test_create_order_from_order_storage_details_with_groups(self):
        _, exchange_manager, trader_inst = await self.init_default()  
        
        order = personal_data.BuyLimitOrder(trader_inst)
        group = exchange_manager.exchange_personal_data.orders_manager.create_group(
            personal_data.OneCancelsTheOtherOrderGroup
        )
        order.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal("70"),
                     quantity=decimal.Decimal("10"),
                     price=decimal.Decimal("70"),
                     group=group)
        order_storage_details = orders_storage._format_order(order, exchange_manager)
    
        pending_groups = {}
        created_order = await personal_data.create_order_from_order_storage_details(
            order_storage_details, exchange_manager, pending_groups
        )
        assert created_order.order_group is group
        assert pending_groups == {group.name: group}
        await self.stop(exchange_manager)

    async def test_create_order_from_order_storage_details_with_chained_orders_with_group(self):
        _, exchange_manager, trader_inst = await self.init_default()
    
        order = personal_data.BuyLimitOrder(trader_inst)
        group_1 = exchange_manager.exchange_personal_data.orders_manager.create_group(
            personal_data.OneCancelsTheOtherOrderGroup
        )
        group_2 = exchange_manager.exchange_personal_data.orders_manager.create_group(
            personal_data.OneCancelsTheOtherOrderGroup
        )
        chained_order_1 = personal_data.BuyLimitOrder(trader_inst)
        chained_order_2 = personal_data.SellLimitOrder(trader_inst)
        chained_order_3 = personal_data.SellLimitOrder(trader_inst)
        for to_update_order in (order, chained_order_1, chained_order_2, chained_order_3):
            to_update_order.update(
                order_type=TraderOrderType.BUY_LIMIT,
                symbol="BTC/USDT",
                current_price=decimal.Decimal("70"),
                quantity=decimal.Decimal("10"),
                price=decimal.Decimal("70")
            )
        chained_order_1.add_to_order_group(group_1)
        chained_order_2.add_to_order_group(group_1)
        chained_order_3.add_to_order_group(group_2)
        await chained_order_1.set_as_chained_order(order, False, {}, False)
        await chained_order_2.set_as_chained_order(order, True, {"plop_1": True, "plop_2": {"hi": 1}}, False)
        order.add_chained_order(chained_order_1)
        order.add_chained_order(chained_order_2)
        await chained_order_3.set_as_chained_order(chained_order_1, False, {}, False)
        chained_order_1.add_chained_order(chained_order_3)
        order_storage_details = orders_storage._format_order(order, exchange_manager)
    
        pending_groups = {}
        created_order = await personal_data.create_order_from_order_storage_details(
            order_storage_details, exchange_manager, pending_groups
        )
        assert pending_groups == {
            group_1.name: group_1,
            group_2.name: group_2,
        }
        chained_orders = created_order.chained_orders
        assert len(chained_orders) == 2
        for chained_order in chained_orders:
            assert chained_order.order_group is group_1
    
        assert chained_orders[0].triggered_by is created_order
        assert chained_orders[0].has_been_bundled is False
        assert chained_orders[0].exchange_creation_params == {}
        assert chained_orders[1].triggered_by is created_order
        assert chained_orders[1].has_been_bundled is True
        assert chained_orders[1].exchange_creation_params == {"plop_1": True, "plop_2": {"hi": 1}}
        assert chained_orders[1].chained_orders == []
        second_level_chained_orders = chained_orders[0].chained_orders
        assert len(second_level_chained_orders) == 1
        assert second_level_chained_orders[0].order_group is group_2
        assert second_level_chained_orders[0].chained_orders == []
        assert second_level_chained_orders[0].triggered_by is chained_orders[0]
        assert second_level_chained_orders[0].has_been_bundled is False
        assert second_level_chained_orders[0].exchange_creation_params == {}
        await self.stop(exchange_manager)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.personal_data.portfolios.assets.future_asset as future_asset
import octobot_trading.personal_data.portfolios.portfolio as portfolio_class
import octobot_trading.constants as constants


class FuturePortfolio(portfolio_class.Portfolio):
    def create_currency_asset(self, currency, available=constants.ZERO, total=constants.ZERO):
        return future_asset.FutureAsset(name=currency, available=available, total=total)

    def update_portfolio_data_from_order(self, order):
        """
        Nothing to do on futures, the portfolio will be updated once the position is updated
        from the order.
        :param order: the order that updated the portfolio
        """

    def update_portfolio_data_from_withdrawal(self, amount, currency):
        """
        Call update_portfolio_data for order currency and market
        :param amount: the withdrawal amount
        :param currency: the withdrawal currency
        """
        self._update_future_portfolio_data(currency, wallet_value=-amount)

    def update_portfolio_data_from_position_size_update(self,
                                                        position,
                                                        realized_pnl_update,
                                                        size_update,
                                                        margin_update,
                                                        has_increased_position_size):
        """
        Update portfolio data from position size update
        :param position: the position updated instance
        :param realized_pnl_update: the position realised pnl update
        :param size_update: the position size update
        :param margin_update: the position margin update
        :param has_increased_position_size: if the update increased position size
        """
        # when the update comes from a closed order, simultaneously update order and position margin
        pair_future_contract = position.symbol_contract

        # get real update quantity from position's contract leverage
        real_update_quantity = decimal.Decimal(size_update / pair_future_contract.current_leverage).copy_abs()

        try:
            # When inverse contract, decrease a currency market equivalent quantity from currency balance
            if pair_future_contract.is_inverse_contract():
                order_margin_update_quantity = real_update_quantity / position.mark_price
                self._update_future_portfolio_data(
                    position.currency,
                    wallet_value=realized_pnl_update,
                    position_margin_value=margin_update,
                    order_margin_value=-order_margin_update_quantity if has_increased_position_size else constants.ZERO)

            # When non-inverse contract, decrease directly market quantity
            else:
                # decrease market quantity from market available balance
                order_margin_update_quantity = real_update_quantity * position.mark_price
                self._update_future_portfolio_data(
                    position.market,
                    wallet_value=realized_pnl_update,
                    position_margin_value=margin_update,
                    order_margin_value=-order_margin_update_quantity if has_increased_position_size else constants.ZERO)
        except (decimal.DivisionByZero, decimal.InvalidOperation) as e:
            self.logger.error(f"Failed to update from position size update : {position} ({e})")

    def update_portfolio_available_from_order(self, order, is_new_order=True):
        """
        Realise portfolio availability update
        :param order: the order that triggers the portfolio update
        :param is_new_order: True when the order is being created
        """
        pair_future_contract = order.exchange_manager.exchange.get_pair_future_contract(order.symbol)
        position_instance = order.exchange_manager.exchange_personal_data.positions_manager. \
            get_order_position(order, contract=pair_future_contract)

        if not position_instance.is_order_increasing_size(order):
            return  # decreasing position size order are not impacting availability

        try:
            real_order_quantity = (order.origin_quantity / pair_future_contract.current_leverage
                                   * (constants.ONE if is_new_order else -constants.ONE))

            # When inverse contract, decrease a currency market equivalent quantity from currency balance
            if pair_future_contract.is_inverse_contract():
                self._update_future_portfolio_data(order.currency,
                                                   wallet_value=constants.ZERO,
                                                   position_margin_value=constants.ZERO,
                                                   order_margin_value=real_order_quantity / order.origin_price)
            # When non-inverse contract, decrease directly market quantity
            else:
                # decrease market quantity from market available balance
                self._update_future_portfolio_data(order.market,
                                                   wallet_value=constants.ZERO,
                                                   position_margin_value=constants.ZERO,
                                                   order_margin_value=real_order_quantity * order.origin_price)
        except (decimal.DivisionByZero, decimal.InvalidOperation) as e:
            self.logger.error(f"Failed to update available from order : {order} ({e})")

    def update_portfolio_from_funding(self, position, funding_rate):
        """
        Update portfolio from funding
        :param position: the perpetual position
        :param funding_rate: the funding rate
        """
        try:
            funding_fee = position.value * funding_rate
            # When inverse contract, decrease a currency market equivalent quantity from currency balance
            if position.symbol_contract.is_inverse_contract():
                self._update_future_portfolio_data(position.currency, wallet_value=-funding_fee)
            # When non-inverse contract, decrease directly market quantity
            else:
                self._update_future_portfolio_data(position.market, wallet_value=-funding_fee)
            self.logger.debug(f"Updated position from funding fees ({str(-funding_fee)}) : {position}")
        except (decimal.DivisionByZero, decimal.InvalidOperation) as e:
            self.logger.error(f"Failed to update from funding : {position} ({e})")

    def update_portfolio_from_pnl(self, position):
        """
        Updates the portfolio from a Position PNL update
        TODO: manage portfolio PNL update when using cross margin
        :param position: the updating position instance
        """
        if position.symbol_contract.is_isolated():
            self.get_currency_portfolio(
                currency=position.currency if position.symbol_contract.is_inverse_contract() else position.market). \
                set_unrealized_pnl(position.unrealized_pnl)

    def _update_future_portfolio_data(self, currency,
                                      wallet_value=constants.ZERO,
                                      position_margin_value=constants.ZERO,
                                      order_margin_value=constants.ZERO,
                                      unrealized_pnl_value=constants.ZERO,
                                      initial_margin_value=constants.ZERO,
                                      replace_value=False):
        """
        Set new currency quantity in the portfolio
        :param currency: the currency to update
        :param wallet_value: the wallet balance value
        :param initial_margin_value: the initial margin value
        :param position_margin_value: the position margin value
        :param order_margin_value: the order margin value
        :param unrealized_pnl_value: the unrealized pnl value
        :param replace_value: when True replace the current value instead of updating it
        :return: True if updated
        """
        try:
            if replace_value:
                return self.portfolio[currency].set(total=wallet_value,
                                                    available=order_margin_value,
                                                    initial_margin=initial_margin_value,
                                                    position_margin=position_margin_value,
                                                    unrealized_pnl=unrealized_pnl_value)
            return self.portfolio[currency].update(total=wallet_value,
                                                   available=order_margin_value,
                                                   position_margin=position_margin_value,
                                                   initial_margin=initial_margin_value,
                                                   unrealized_pnl=unrealized_pnl_value)
        except KeyError:
            self.portfolio[currency] = self.create_currency_asset(currency=currency,
                                                                  available=order_margin_value,
                                                                  total=wallet_value)
            return True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import asyncio

import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_commons.databases as commons_databases

import octobot_backtesting.api as backtesting_api
import octobot_trading.storage.abstract_storage as abstract_storage
import octobot_trading.util as util
import octobot_trading.constants as constants


class CandlesStorage(abstract_storage.AbstractStorage):
    IS_LIVE_CONSUMER = False
    IS_HISTORICAL = False
    HISTORY_TABLE = commons_enums.DBTables.CANDLES_SOURCE.value
    ENABLE_LIVE_CANDLES_STORAGE = constants.ENABLE_LIVE_CANDLES_STORAGE

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_task = None
        self._init_timeout = 5 * commons_constants.MINUTE_TO_SECONDS
        self.enabled = (self.exchange_manager is not None and self.exchange_manager.is_backtesting) \
            or self.ENABLE_LIVE_CANDLES_STORAGE

    async def on_start(self):
        self._init_task = asyncio.create_task(self._store_candles_when_available())

    async def stop(self, **kwargs):
        if self._init_task is not None and not self._init_task.done():
            self._init_task.cancel()
        await super().stop(**kwargs)

    async def _store_candles_when_available(self):
        await util.wait_for_topic_init(self.exchange_manager, self._init_timeout,
                                       commons_enums.InitializationEventExchangeTopics.CANDLES.value)
        await self.store_candles()

    @abstract_storage.AbstractStorage.hard_reset_and_retry_if_necessary
    async def store_candles(self):
        if not self.enabled:
            return
        for symbol in self.exchange_manager.exchange_config.traded_symbol_pairs:
            symbol_db = self._get_db(symbol)
            await symbol_db.delete(self.HISTORY_TABLE, None)
            for time_frame in self.exchange_manager.exchange_config.get_relevant_time_frames():
                await self._store_candles_if_necessary(symbol, time_frame.value, symbol_db)
            await symbol_db.flush()

    async def _store_candles_if_necessary(self, symbol, time_frame, symbol_db):
        candles_data = {
            commons_enums.DBRows.TIME_FRAME.value: time_frame,
            commons_enums.DBRows.VALUE.value: backtesting_api.get_data_file_from_importers(
                self.exchange_manager.exchange.connector.exchange_importers, symbol,
                commons_enums.TimeFrames(time_frame)
            )
            if self.exchange_manager.is_backtesting else commons_constants.LOCAL_BOT_DATA,
            commons_enums.DisplayedElementTypes.CHART.value: self.plot_settings.chart
        }
        if (not await symbol_db.contains_row(
                self.HISTORY_TABLE,
                {
                    commons_enums.DBRows.TIME_FRAME.value: time_frame,
                    commons_enums.DBRows.VALUE.value: candles_data[commons_enums.DBRows.VALUE.value],
                })):
            await symbol_db.log(self.HISTORY_TABLE, candles_data)

    async def clear_history(self, flush=True):
        for database in await self._get_all_symbol_dbs():
            await database.delete(self.HISTORY_TABLE, None)
            if flush:
                await database.flush()

    def _get_db(self, symbol):
        return commons_databases.RunDatabasesProvider.instance().get_symbol_db(
            self.exchange_manager.bot_id,
            self.exchange_manager.exchange_name,
            symbol
        )

    async def _get_all_symbol_dbs(self):
        return await commons_databases.RunDatabasesProvider.instance().get_all_symbol_dbs(
            self.exchange_manager.bot_id,
            self.exchange_manager.exchange_name
        )




#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.constants
import octobot_trading.personal_data as personal_data


def get_portfolio(exchange_manager, as_decimal=True) -> dict:
    return format_portfolio(
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio,
        as_decimal
    )


def get_portfolio_historical_values(exchange_manager, currency, time_frame, from_timestamp=None, to_timestamp=None) \
        -> list:
    if exchange_manager.exchange_personal_data.portfolio_manager:
        return exchange_manager.exchange_personal_data.portfolio_manager.get_portfolio_historical_values(
            currency, time_frame, from_timestamp, to_timestamp
        )
    return []


def get_global_portfolio_currencies_values(exchange_managers: list) -> dict:
    currencies_values = {}
    for exchange in exchange_managers:
        this_currency_values = (
            exchange.exchange_personal_data.portfolio_manager
            .portfolio_value_holder.get_current_crypto_currencies_values()
        )
        for currency, value in this_currency_values.items():
            if currency not in currencies_values:
                currencies_values[currency] = value
    return currencies_values


def get_portfolio_currency(exchange_manager, currency) -> personal_data.Asset:
    return exchange_manager.exchange_personal_data.portfolio_manager.portfolio.get_currency_portfolio(currency)


def get_origin_portfolio(exchange_manager, as_decimal=True) -> dict:
    return format_portfolio(
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.origin_portfolio.portfolio,
        as_decimal
    )


def set_simulated_portfolio_initial_config(exchange_manager, portfolio_content):
    if exchange_manager.exchange_personal_data.portfolio_manager:
        exchange_manager.exchange_personal_data.portfolio_manager.set_forced_portfolio_initial_config(
            portfolio_content
        )


def format_portfolio(portfolio, as_decimal) -> dict:
    if as_decimal:
        return portfolio
    return personal_data.portfolio_to_float(portfolio)


def parse_decimal_portfolio(portfolio, as_decimal) -> dict:
    return personal_data.parse_decimal_portfolio(portfolio, as_decimal=as_decimal)


async def refresh_real_trader_portfolio(exchange_manager) -> bool:
    return await exchange_channel.get_chan(octobot_trading.constants.BALANCE_CHANNEL, exchange_manager.id). \
        get_internal_producer(). \
        refresh_real_trader_portfolio(True)


def get_draw_down(exchange_manager) -> decimal.Decimal:
    return personal_data.get_draw_down(exchange_manager)


async def get_coefficient_of_determination(exchange_manager, use_high_instead_of_end_balance=True):
    return await personal_data.get_coefficient_of_determination(exchange_manager,
                                                                use_high_instead_of_end_balance)


def get_usd_like_symbol_from_symbols(currency: str, symbols) -> str:
    return personal_data.ValueConverter.get_usd_like_symbol_from_symbols(currency, symbols)


def can_convert_symbol_to_usd_like(symbol: str) -> bool:
    return personal_data.ValueConverter.can_convert_symbol_to_usd_like(symbol)


def is_usd_like_coin(coin) -> bool:
    return personal_data.ValueConverter.is_usd_like_coin(coin)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.enums as commons_enums
import octobot_commons.databases as commons_databases

import octobot_trading.storage.abstract_storage as abstract_storage
import octobot_trading.storage.util as storage_util


class TransactionsStorage(abstract_storage.AbstractStorage):
    IS_LIVE_CONSUMER = False
    HISTORY_TABLE = commons_enums.DBTables.TRANSACTIONS.value

    @abstract_storage.AbstractStorage.hard_reset_and_retry_if_necessary
    async def _store_history(self):
        transactions = [
            transaction
            for transaction in self.exchange_manager.exchange_personal_data.transactions_manager.transactions.values()
        ]
        y_data = self.plot_settings.y_data or [0] * len(transactions)
        await self._get_db().replace_all(
            self.HISTORY_TABLE,
            [
                _format_transaction(
                    transaction,
                    self.exchange_manager,
                    self.plot_settings.chart,
                    self.plot_settings.x_multiplier,
                    self.plot_settings.kind,
                    self.plot_settings.mode,
                    y_data[index]
                )
                for index, transaction in enumerate(transactions)
            ],
            cache=False,
        )
        await self.trigger_debounced_flush()

    def _get_db(self):
        return commons_databases.RunDatabasesProvider.instance().get_transactions_db(
            self.exchange_manager.bot_id,
            storage_util.get_account_type_suffix_from_exchange_manager(self.exchange_manager),
            self.exchange_manager.exchange_name
        )


def _format_transaction(transaction, exchange_manager, chart, x_multiplier, kind, mode, y_data):
    return {
        commons_enums.DisplayedElementTypes.CHART.value: chart,
        commons_enums.DBRows.SYMBOL.value: transaction.symbol,
        commons_enums.PlotAttributes.X.value: transaction.creation_time * x_multiplier,
        commons_enums.PlotAttributes.TYPE.value: transaction.transaction_type.value,
        commons_enums.PlotAttributes.SIDE.value: transaction.side.value if hasattr(transaction, "side") else None,
        commons_enums.PlotAttributes.Y.value: y_data,
        commons_enums.PlotAttributes.KIND.value: kind,
        commons_enums.PlotAttributes.MODE.value: mode,
        "id": transaction.transaction_id,
        "trading_mode": exchange_manager.trading_modes[0].get_name(),
        "currency": transaction.currency,
        "quantity": float(transaction.quantity) if hasattr(transaction, "quantity") else None,
        "order_id": transaction.order_id if hasattr(transaction, "order_id") else None,
        "funding_rate": float(transaction.funding_rate) if hasattr(transaction, "funding_rate") else None,
        "realised_pnl": float(transaction.realised_pnl) if hasattr(transaction, "realised_pnl") else None,
        "transaction_fee": float(transaction.transaction_fee) if hasattr(transaction, "transaction_fee") else None,
        "closed_quantity": float(transaction.closed_quantity) if hasattr(transaction, "closed_quantity") else None,
        "cumulated_closed_quantity": float(transaction.cumulated_closed_quantity)
        if hasattr(transaction, "cumulated_closed_quantity") else None,
        "first_entry_time": float(transaction.first_entry_time) * x_multiplier
        if hasattr(transaction, "first_entry_time") else None,
        "average_entry_price": float(transaction.average_entry_price)
        if hasattr(transaction, "average_entry_price") else None,
        "average_exit_price": float(transaction.average_exit_price)
        if hasattr(transaction, "average_exit_price") else None,
        "order_exit_price": float(transaction.order_exit_price)
        if hasattr(transaction, "order_exit_price") else None,
        "leverage": float(transaction.leverage) if hasattr(transaction, "leverage") else None,
        "trigger_source": transaction.trigger_source.value if hasattr(transaction, "trigger_source") else None,
    }



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

from tests import event_loop
import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums

from octobot_trading.api.exchange import cancel_ccxt_throttle_task

from tests.exchanges import simulated_exchange_manager, simulated_trader
from tests.personal_data.trades import create_executed_trade


SYMBOL = "BTC/USDT"


def test_aggregate_trades_by_exchange_order_id(simulated_trader):
    _, _, trader = simulated_trader
    t1 = create_executed_trade(
        trader, enums.TradeOrderSide.BUY, 1, decimal.Decimal(10), decimal.Decimal(1500), SYMBOL, _get_fees("BTC", 0.1)
    )
    t2 = create_executed_trade(
        trader, enums.TradeOrderSide.BUY, 3, decimal.Decimal(13), decimal.Decimal(1605), SYMBOL, _get_fees("BTC", 0.5)
    )
    t3 = create_executed_trade(
        trader, enums.TradeOrderSide.BUY, 7, decimal.Decimal(7), decimal.Decimal(1400), SYMBOL, _get_fees("BTC", 1.4)
    )
    aggregated = personal_data.aggregate_trades_by_exchange_order_id([t1, t2, t3])
    # aggregated all trades together as all of their exchange_order_id is None
    assert list(aggregated) == [None]
    assert aggregated[None].trader is trader
    assert aggregated[None].symbol == t1.symbol
    assert aggregated[None].side is t1.side
    assert aggregated[None].executed_price == decimal.Decimal("1522.166666666666666666666667")
    assert aggregated[None].executed_quantity == sum(t.executed_quantity for t in (t1, t2, t3))
    assert aggregated[None].total_cost == sum(t.total_cost for t in (t1, t2, t3))
    assert aggregated[None].fee[enums.FeePropertyColumns.COST.value] == \
       sum(t.fee[enums.FeePropertyColumns.COST.value] for t in (t1, t2, t3))
    assert aggregated[None].executed_time == 7

    # set exchange_order_id
    t1.exchange_order_id = "1"
    t2.exchange_order_id = "1"
    t3.exchange_order_id = "2"
    aggregated = personal_data.aggregate_trades_by_exchange_order_id([t1, t2, t3])
    # aggregated all trades together as all of their exchange_order_id is None
    assert list(aggregated) == ["1", "2"]
    assert aggregated["1"].trader is trader
    assert aggregated["1"].symbol == t1.symbol
    assert aggregated["1"].side is t1.side
    assert aggregated["1"].executed_price == decimal.Decimal("1559.347826086956521739130435")
    assert aggregated["1"].executed_quantity == sum(t.executed_quantity for t in (t1, t2))
    assert aggregated["1"].total_cost == sum(t.total_cost for t in (t1, t2))
    assert aggregated["1"].fee[enums.FeePropertyColumns.COST.value] == \
       sum(t.fee[enums.FeePropertyColumns.COST.value] for t in (t1, t2))
    assert aggregated["1"].executed_time == 3

    assert aggregated["2"].executed_price == decimal.Decimal(1400)
    assert aggregated["2"].executed_quantity == decimal.Decimal(7)
    assert aggregated["2"].total_cost == t3.total_cost
    assert aggregated["2"].fee == _get_fees("BTC", 1.4)
    assert aggregated["2"].executed_time == 7


def _get_fees(currency, value):
    return {
        enums.FeePropertyColumns.CURRENCY.value: currency,
        enums.FeePropertyColumns.COST.value: decimal.Decimal(str(value))
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import pytest
import tinydb.table

import octobot_trading.personal_data as personal_data


def test_create_historical_asset_value_from_dict_invalid_input():
    input_1 = {
        personal_data.HistoricalAssetValue.TIMESTAMP_KEY: 11,
        f"{personal_data.HistoricalAssetValue.VALUES_KEY}_invalid": {
            "BTC": 1,
            "USD": 222
        },
    }
    with pytest.raises(KeyError):
        personal_data.create_historical_asset_value_from_dict_like_object(personal_data.HistoricalAssetValue, input_1)
    input_2 = {
        f"{personal_data.HistoricalAssetValue.VALUES_KEY}_invalid": {
            "BTC": 1,
            "USD": 222
        },
    }
    with pytest.raises(KeyError):
        personal_data.create_historical_asset_value_from_dict_like_object(personal_data.HistoricalAssetValue, input_2)


def test_create_historical_asset_value_from_dict_valid_input():
    input_1 = {
        personal_data.HistoricalAssetValue.TIMESTAMP_KEY: 11,
        personal_data.HistoricalAssetValue.VALUES_KEY: {
            "BTC": 1,
            "USD": 222
        },
    }
    historical_asset = personal_data.create_historical_asset_value_from_dict_like_object(personal_data.HistoricalAssetValue,
                                                                                         input_1)
    assert historical_asset.get_timestamp() == 11
    assert list(historical_asset.get_currencies()) == ["BTC", "USD"]
    assert historical_asset.get("BTC") == decimal.Decimal(1)
    assert historical_asset.get("USD") == decimal.Decimal(222)
    assert historical_asset.to_dict() == input_1

    input_2 = {
        personal_data.HistoricalAssetValue.TIMESTAMP_KEY: 1111111111111.222,
        personal_data.HistoricalAssetValue.VALUES_KEY: {},
    }
    assert personal_data.create_historical_asset_value_from_dict_like_object(
        personal_data.HistoricalAssetValue,
        input_2
    ).to_dict() == input_2

    # input can be a tinydb.table.Document
    assert personal_data.create_historical_asset_value_from_dict_like_object(
        personal_data.HistoricalAssetValue,
        tinydb.table.Document(input_2, 1)
    ).to_dict() == input_2

    input_3 = {
        personal_data.HistoricalAssetValue.TIMESTAMP_KEY: 0,
        personal_data.HistoricalAssetValue.VALUES_KEY: {
            "BTC": 1.7777777777777,
            "XXXXX": 0
        },
    }
    assert personal_data.create_historical_asset_value_from_dict_like_object(
        personal_data.HistoricalAssetValue,
        input_3
    ).to_dict() == input_3
